# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'define_external_forces.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ExternalForces(object):
    def setupUi(self, ExternalForces):
        ExternalForces.setObjectName("ExternalForces")
        ExternalForces.resize(393, 288)
        self.verticalLayout = QtWidgets.QVBoxLayout(ExternalForces)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.upGroupBox = QtWidgets.QGroupBox(ExternalForces)
        self.upGroupBox.setObjectName("upGroupBox")
        self.formLayout = QtWidgets.QFormLayout(self.upGroupBox)
        self.formLayout.setObjectName("formLayout")
        self.fractionOfSpaceLabel = QtWidgets.QLabel(self.upGroupBox)
        self.fractionOfSpaceLabel.setObjectName("fractionOfSpaceLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.fractionOfSpaceLabel)
        self.upFractionOfSpaceDoubleSpinBox = QtWidgets.QDoubleSpinBox(self.upGroupBox)
        self.upFractionOfSpaceDoubleSpinBox.setDecimals(4)
        self.upFractionOfSpaceDoubleSpinBox.setMaximum(1.0)
        self.upFractionOfSpaceDoubleSpinBox.setSingleStep(0.01)
        self.upFractionOfSpaceDoubleSpinBox.setProperty("value", 0.05)
        self.upFractionOfSpaceDoubleSpinBox.setObjectName("upFractionOfSpaceDoubleSpinBox")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.upFractionOfSpaceDoubleSpinBox)
        self.lowerLimitMLabel = QtWidgets.QLabel(self.upGroupBox)
        self.lowerLimitMLabel.setObjectName("lowerLimitMLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lowerLimitMLabel)
        self.lowerLimitMLineEdit = QtWidgets.QLineEdit(self.upGroupBox)
        self.lowerLimitMLineEdit.setObjectName("lowerLimitMLineEdit")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lowerLimitMLineEdit)
        self.xUpBtn = QtWidgets.QToolButton(self.upGroupBox)
        self.xUpBtn.setObjectName("xUpBtn")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.xUpBtn)
        self.applyXUp = QtWidgets.QCheckBox(self.upGroupBox)
        self.applyXUp.setObjectName("applyXUp")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.applyXUp)
        self.yUpBtn = QtWidgets.QToolButton(self.upGroupBox)
        self.yUpBtn.setObjectName("yUpBtn")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.yUpBtn)
        self.applyYUp = QtWidgets.QCheckBox(self.upGroupBox)
        self.applyYUp.setObjectName("applyYUp")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.applyYUp)
        self.verticalLayout.addWidget(self.upGroupBox)
        self.lowGroupBox = QtWidgets.QGroupBox(ExternalForces)
        self.lowGroupBox.setObjectName("lowGroupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.lowGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.fractionOfSpaceLabel_2 = QtWidgets.QLabel(self.lowGroupBox)
        self.fractionOfSpaceLabel_2.setObjectName("fractionOfSpaceLabel_2")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.fractionOfSpaceLabel_2)
        self.lowFractionOfSpaceDoubleSpinBox = QtWidgets.QDoubleSpinBox(self.lowGroupBox)
        self.lowFractionOfSpaceDoubleSpinBox.setDecimals(4)
        self.lowFractionOfSpaceDoubleSpinBox.setMaximum(1.0)
        self.lowFractionOfSpaceDoubleSpinBox.setSingleStep(0.01)
        self.lowFractionOfSpaceDoubleSpinBox.setProperty("value", 0.05)
        self.lowFractionOfSpaceDoubleSpinBox.setObjectName("lowFractionOfSpaceDoubleSpinBox")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lowFractionOfSpaceDoubleSpinBox)
        self.upperLimitMLabel = QtWidgets.QLabel(self.lowGroupBox)
        self.upperLimitMLabel.setObjectName("upperLimitMLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.upperLimitMLabel)
        self.upperLimitMLineEdit = QtWidgets.QLineEdit(self.lowGroupBox)
        self.upperLimitMLineEdit.setObjectName("upperLimitMLineEdit")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.upperLimitMLineEdit)
        self.blockLabel = QtWidgets.QLabel(self.lowGroupBox)
        self.blockLabel.setObjectName("blockLabel")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.blockLabel)
        self.blockCheckBox = QtWidgets.QCheckBox(self.lowGroupBox)
        self.blockCheckBox.setObjectName("blockCheckBox")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.blockCheckBox)
        self.verticalLayout_3.addLayout(self.formLayout_2)
        self.verticalLayout.addWidget(self.lowGroupBox)

        self.retranslateUi(ExternalForces)
        QtCore.QMetaObject.connectSlotsByName(ExternalForces)

    def retranslateUi(self, ExternalForces):
        _translate = QtCore.QCoreApplication.translate
        self.upGroupBox.setTitle(_translate("ExternalForces", "Upper zone (y+)"))
        self.fractionOfSpaceLabel.setText(_translate("ExternalForces", "Fraction of space"))
        self.lowerLimitMLabel.setText(_translate("ExternalForces", "Lower limit (m)"))
        self.xUpBtn.setText(_translate("ExternalForces", "force x component"))
        self.applyXUp.setText(_translate("ExternalForces", "Apply"))
        self.yUpBtn.setText(_translate("ExternalForces", "force y component"))
        self.applyYUp.setText(_translate("ExternalForces", "Apply"))
        self.lowGroupBox.setTitle(_translate("ExternalForces", "Lower zone (y-)"))
        self.fractionOfSpaceLabel_2.setText(_translate("ExternalForces", "Fraction of space"))
        self.upperLimitMLabel.setText(_translate("ExternalForces", "Upper limit (m)"))
        self.blockLabel.setText(_translate("ExternalForces", "Block"))

