# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# datreant.tests

"""
datreant.tests --- unit tests for datreant
=============================================================

"""
# Bring some often used objects into the current namespace
from . import test_collections
from . import test_locks
from . import test_treants
